/*
	MSharpen plugin for Avisynth -- sharpens important edges in the
	image without amplifying noise. The filter operates by detecting
	edges and then applying a sharpening kernel only in the edge areas.

	Copyright (C) 2002 Donald A. Graft, All Rights Reserved

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Donald Graft
	neuron2@attbi.com.
*/

#include "internal.h"
#include "windows.h"
#include "stdio.h"

class MSharpen : public GenericVideoFilter {
	bool mask, highq, debug;
	int threshold, strength, invstrength;
public:
    MSharpen(PClip _child, int _threshold, int _strength, bool _highq, bool _mask, 
		 bool _debug, IScriptEnvironment* env) :
		 GenericVideoFilter(_child), threshold(_threshold), strength(_strength),
		 highq(_highq), mask(_mask), debug(_debug)
	{
		if (!vi.IsYUY2() && !vi.IsRGB32() && (!vi.IsPlanar()))
			env->ThrowError("MSharpen: YUY2, RGB32 or planar (YV12 for instance) color space required");
		if (strength < 0 || strength > 255)
			env->ThrowError("MSharpen: strength out of range (0-255)");
		invstrength = 255 - strength;
		if (debug)
		{
			char b[80];
			sprintf(b, "MSharpen %s by Donald Graft, Copyright 2003\n", "1.10 beta 2");
			OutputDebugString(b);
		}

    if (vi.IsPlanar() && (!(env->GetCPUFlags() & CPUF_INTEGER_SSE)))
      env->ThrowError("MSharpen: Planar image processing requires an ISSE capable processor (Athlon, P3 or later)");
    if (vi.IsPlanar() && (vi.width&3))
      env->ThrowError("MSharpen: Planar image processing requires a width divisible by 4");

	}
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
		void blur_plane(PVideoFrame src, PVideoFrame blur, unsigned char *blurp, int plane);
    void detect_edges(PVideoFrame src, PVideoFrame dst, unsigned char *dstp, int plane);
    void detect_edges_HiQ(PVideoFrame src, PVideoFrame dst, unsigned char *dstp, int plane);
    void apply_filter(PVideoFrame src, PVideoFrame blur, PVideoFrame dst, unsigned char *dstp, int plane);
};

PVideoFrame __stdcall MSharpen::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame blur = env->NewVideoFrame(vi);
	PVideoFrame dst = env->NewVideoFrame(vi);
	if (vi.IsPlanar())
	{
		for (int i=0;i<3;i++)
		{
			int plane=0;

			switch (i)
			{
				case 0:
					plane = PLANAR_Y;
					break;
				case 1:
					plane = PLANAR_U;
					break;
				case 2:
					plane = PLANAR_V;
					break;
			}
			unsigned char *blurp = blur->GetWritePtr(plane);
			blur_plane(src, blur, blurp ,plane);

			unsigned char *dstp = dst->GetWritePtr(plane);
			detect_edges(blur, dst, dstp, plane);
			if (highq == true)
				detect_edges_HiQ(blur, dst, dstp, plane);

			if (!mask) apply_filter(src, blur, dst, dstp, plane);
		}
		return dst;
	}

	PVideoFrame work = env->NewVideoFrame(vi);

	const unsigned char *srcp = src->GetReadPtr();
	const unsigned char *srcp_saved = srcp;
	int src_pitch = src->GetPitch();
	const unsigned char *srcpp;
	const unsigned char *srcpn;

	unsigned char *blurp = blur->GetWritePtr();
	unsigned char *blurp_saved = blurp;
	int blur_pitch = blur->GetPitch();
	const unsigned char *blurpn;

	unsigned char *workp = work->GetWritePtr();
	unsigned char *workp_saved = workp;

	unsigned char *dstp = dst->GetWritePtr();
	unsigned char *dstp_saved = dstp;

	int row_size = blur->GetRowSize();
	int height = blur->GetHeight();
	int x, y, max;

	if (vi.IsYUY2())
	{
		int y11, y12, y13, y14, y21, y22, y23, y24, u1, u2, u3, u4, v1, v2, v3, v4;

		/* Blur the source image prior to detail detection. Separate
		   dimensions for speed. */
		/* Vertical blur. */
		srcpp = srcp_saved;
		srcp  = srcpp + src_pitch;
		srcpn = srcp + src_pitch;
		workp = workp_saved + blur_pitch;
		for (y = 1; y < height - 1; y++)
		{
			for (x = 0; x < row_size; x++)
			{
				workp[x] = (srcpp[x] + srcp[x] + srcpn[x]) / 3;
			}
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			workp += blur_pitch;
		}

		/* Horizontal blur. */
		workp  = workp_saved;
		blurp  = blurp_saved;
		for (y = 0; y < height; y++)
		{
			/* Just do luma. Horizontal chroma blurring is a
			   nightmare, this is much faster, and we don't need
			   it anyway. */
			for (x = 2; x < row_size - 2; x+=2)
			{
				blurp[x] = (workp[x-2] + workp[x] + workp[x+2]) / 3;
				blurp[x+1] = workp[x+1];
			}
			workp += blur_pitch;
			blurp += blur_pitch;
		}

		/* Fix up blur frame borders. */
		srcp = srcp_saved;
		blurp = blurp_saved;
		memcpy(blurp, srcp, row_size);
		memcpy(blurp + (height-1)*blur_pitch, srcp + (height-1)*src_pitch, row_size);
		for (y = 0; y < height; y++)
		{
			*((unsigned int *)(&blurp[0])) = *((unsigned int *)(&srcp[0]));
			*((unsigned int *)(&blurp[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
			srcp += src_pitch;
			blurp += blur_pitch;
		}

		/* Diagonal detail detection. */
		blurp = (unsigned char *) blurp_saved;
		blurpn = blurp + blur_pitch;
		dstp = dstp_saved;
		for (y = 0; y < height - 1; y++) 
		{
			y11 = blurp[0];
			u1  = blurp[1];
			y21 = blurp[2];
			v1  = blurp[3];
			y13 = blurpn[0];
			u3  = blurpn[1];
			y23 = blurpn[2];
			v3  = blurpn[3];
			for (x = 0; x < row_size - 4; x+=4)
			{
				y12 = blurp[x+4];
				u2  = blurp[x+5];
				y22 = blurp[x+6];
				v2  = blurp[x+7];
				y14 = blurpn[x+4];
				u4  = blurpn[x+5];
				y24 = blurpn[x+6];
				v4  = blurpn[x+7];
				if ((abs(y11 - y14) >= threshold) || (abs(y12 - y13) >= threshold))
				{
					*((unsigned short *)(&dstp[x])) = 0x80eb;
				}
				else
				{
					*((unsigned short *)(&dstp[x])) = 0x8010;
				}
				if ((abs(y21 - y24) >= threshold) || (abs(y22 - y23) >= threshold))
				{
					*((unsigned short *)(&dstp[x+2])) = 0x80eb;
				}
				else
				{
					*((unsigned short *)(&dstp[x+2])) = 0x8010;
				}
				if ((abs(u1 - u4) >= threshold) || (abs(v1 - v4) >= threshold) ||
					(abs(u2 - u3) >= threshold) || (abs(v2 - v3) >= threshold))
				{
					*((unsigned short *)(&dstp[x]))   = 0x80eb;
					*((unsigned short *)(&dstp[x+2])) = 0x80eb;
				}
				y11 = y12; y21 = y22;
				y13 = y14; y23 = y24;
				u1 = u2; u3 = u4;
				v1 = v2; v3 = v4;
			}
			dstp += blur_pitch;
			blurp += blur_pitch;
			blurpn += blur_pitch;
		}

		if (highq == true)
		{
			/* Vertical detail detection. */
			for (x = 0; x < row_size; x+=4)
			{
 				blurp = blurp_saved;
				blurpn = blurp + blur_pitch;
				dstp = dstp_saved;
				y11 = blurp[x];
				u1  = blurp[x+1];
				y21 = blurp[x+2];
				v1  = blurp[x+3];
				for (y = 0; y < height - 1; y++)
				{
					y12 = blurpn[x];
					u2  = blurpn[x+1];
					y22 = blurpn[x+2];
					v2  = blurpn[x+3];
					if (abs(y11 - y12) >= threshold)
					{
						*((unsigned short *)(&dstp[x])) = 0x80eb;
					}
					if (abs(y21 - y22) >= threshold)
					{
						*((unsigned short *)(&dstp[x+2])) = 0x80eb;
					}
					if (abs(u1 - u2) >= threshold || abs(v1 - v2) >= threshold)
					{
						*((unsigned short *)(&dstp[x]))   = 0x80eb;
						*((unsigned short *)(&dstp[x+2])) = 0x80eb;
					}
					y11 = y12;
					y21 = y22;
					u1 = u2;
					v1 = v2;
					dstp += blur_pitch;
					blurp += blur_pitch;
					blurpn += blur_pitch;
				}
			}

			/* Horizontal detail detection. */
			blurp = blurp_saved;
			dstp = dstp_saved;
			for (y = 0; y < height; y++)
			{
				y11 = blurp[0];
				u1  = blurp[1];
				y21 = blurp[2];
				v1  = blurp[3];
				for (x = 0; x < row_size - 4; x+=4)
				{
					y12 = blurp[x+4];
					u2  = blurp[x+5];
					y22 = blurp[x+6];
					v2  = blurp[x+7];
					if (abs(y11 - y12) >= threshold)
					{
						*((unsigned short *)(&dstp[x])) = 0x80eb;
					}
					if (abs(y21 - y22) >= threshold)
					{
						*((unsigned short *)(&dstp[x+2])) = 0x80eb;
					}
					if (abs(u1 - u2) >= threshold || abs(v1 - v2) >= threshold)
					{
						*((unsigned short *)(&dstp[x]))   = 0x80eb;
						*((unsigned short *)(&dstp[x+2])) = 0x80eb;
					}
					y11 = y12;
					y21 = y22;
					u1 = u2;
					v1 = v2;
				}
				dstp += blur_pitch;
				blurp += blur_pitch;
			}
		}

		/* Fix up detail map borders. */
		dstp = dstp_saved + (height-1)*blur_pitch;
		for (x = 0; x < row_size / 4; x++) ((unsigned int *)(dstp))[x] = 0x80108010;
		dstp = dstp_saved;
		for (y = 0; y < height; y++)
		{
			((unsigned int *)(dstp))[row_size/4-1] = 0x80108010;
			dstp += blur_pitch;
		}

		if (mask == true) return dst;

		/* Fix up output frame borders. */
		srcp = srcp_saved;
		workp = workp_saved;
		memcpy(workp, srcp, row_size);
		memcpy(workp + (height-1)*blur_pitch, srcp + (height-1)*src_pitch, row_size);
		for (y = 0; y < height; y++)
		{
			*((unsigned int *)(&workp[0])) = *((unsigned int *)(&srcp[0]));
			*((unsigned int *)(&workp[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
			srcp += src_pitch;
			workp += blur_pitch;
		}

		/* Now sharpen the edge areas and we're done! */
		srcp = srcp_saved + src_pitch;
 		dstp = dstp_saved + blur_pitch;
		workp = workp_saved + blur_pitch;
		blurp = blurp_saved + blur_pitch;
		for (y = 1; y < height - 1; y++)
		{
			for (x = 2; x < row_size - 2; x+=2)
			{
				/* Don't sharpen chroma. */
				workp[x+1] = srcp[x+1];
				/* Sharpen luma. */
				if (dstp[x] == 0xeb)
				{
					y14 = 4 * (int)srcp[x] - 3 * blurp[x];
					if (y14 < 16) y14 = 16;
					if (y14 > 235) y14 = 235;
					workp[x] = (strength * y14 + invstrength * srcp[x]) >> 8;
				}
				else
				{
					workp[x] = srcp[x];
				}
			}
			srcp += src_pitch;
			dstp += blur_pitch;
			workp += blur_pitch;
			blurp += blur_pitch;
		}
		return work;
	}
	else if (vi.IsRGB())
	{
		int r1, r2, r3, r4, g1, g2, g3, g4, b1, b2, b3, b4;

		/* Blur the source image prior to detail detection. Separate
		   dimensions for speed. */
		/* Vertical. */
		srcpp = srcp_saved;
		srcp  = srcpp + src_pitch;
		srcpn = srcp + src_pitch;
		workp = workp_saved + blur_pitch;
		for (y = 1; y < height - 1; y++)
		{
			for (x = 0; x < row_size; x++)
			{
				workp[x] = (srcpp[x] + srcp[x] + srcpn[x]) / 3;
			}
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			workp += blur_pitch;
		}
		/* Horizontal. */
		workp  = workp_saved;
		blurp  = blurp_saved;
		for (y = 0; y < height; y++)
		{
			for (x = 4; x < row_size - 4; x++)
			{
				blurp[x] = (workp[x-4] + workp[x] + workp[x+4]) / 3;
			}
			workp += blur_pitch;
			blurp += blur_pitch;
		}

		/* Fix up blur frame borders. */
		srcp = srcp_saved;
		blurp = blurp_saved;
		memcpy(blurp, srcp, row_size);
		memcpy(blurp + (height-1)*blur_pitch, srcp + (height-1)*src_pitch, row_size);
		for (y = 0; y < height; y++)
		{
			*((unsigned int *)(&blurp[0])) = *((unsigned int *)(&srcp[0]));
			*((unsigned int *)(&blurp[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
			srcp += src_pitch;
			blurp += blur_pitch;
		}

		/* Diagonal detail detection. */
		blurp = (unsigned char *) blurp_saved;
		blurpn = blurp + blur_pitch;
		dstp = dstp_saved;
		for (y = 0; y < height - 1; y++) 
		{
			b1 = blurp[0];
			g1 = blurp[1];
			r1 = blurp[2];
			b3 = blurpn[0];
			g3 = blurpn[1];
			r3 = blurpn[2];
			for (x = 0; x < row_size - 4; x+=4)
			{
				b2 = blurp[x+4];
				g2 = blurp[x+5];
				r2 = blurp[x+6];
				b4 = blurpn[x+4];
				g4 = blurpn[x+5];
				r4 = blurpn[x+6];
				if ((abs(b1 - b4) >= threshold) || (abs(g1 - g4) >= threshold) || (abs(r1 - r4) >= threshold) ||
					(abs(b2 - b3) >= threshold) || (abs(g2 - g3) >= threshold) || (abs(g2 - g3) >= threshold))
				{
					*((unsigned int *)(&dstp[x])) = 0xffffffff;
				}
				else
				{
					*((unsigned int *)(&dstp[x])) = 0x0;
				}
				b1 = b2; b3 = b4;
				g1 = g2; g3 = g4;
				r1 = r2; r3 = r4;
			}
			dstp += blur_pitch;
			blurp += blur_pitch;
			blurpn += blur_pitch;
		}

		if (highq == true)
		{
			/* Vertical detail detection. */
			for (x = 0; x < row_size; x+=4)
			{
 				blurp = blurp_saved;
				blurpn = blurp + blur_pitch;
				dstp = dstp_saved;
				b1 = blurp[x];
				g1 = blurp[x+1];
				r1 = blurp[x+2];
				for (y = 0; y < height - 1; y++)
				{
					b2 = blurpn[x];
					g2 = blurpn[x+1];
					r2 = blurpn[x+2];
					if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold)
					{
						*((unsigned int *)(&dstp[x])) = 0xffffffff;
					}
					b1 = b2;
					g1 = g2;
					r1 = r2;
					dstp += blur_pitch;
					blurp += blur_pitch;
					blurpn += blur_pitch;
				}
			}

			/* Horizontal detail detection. */
			blurp = blurp_saved;
			dstp = dstp_saved;
			for (y = 0; y < height; y++)
			{
				b1 = blurp[0];
				g1 = blurp[1];
				r1 = blurp[2];
				for (x = 0; x < row_size - 4; x+=4)
				{
					b2 = blurp[x+4];
					g2 = blurp[x+5];
					r2 = blurp[x+6];
					if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold)
					{
						*((unsigned int *)(&dstp[x])) = 0xffffffff;
					}
					b1 = b2;
					g1 = g2;
					r1 = r2;
				}
				dstp += blur_pitch;
				blurp += blur_pitch;
			}
		}

		/* Fix up detail map borders. */
		memset(dstp_saved + (height-1)*blur_pitch, 0, row_size);
		dstp = dstp_saved;
		for (y = 0; y < height; y++)
		{
			*((unsigned int *)(&dstp[row_size-4])) = 0;
			dstp += blur_pitch;
		}

		if (mask == true) return dst;

		/* Fix up output frame borders. */
		srcp = srcp_saved;
		workp = workp_saved;
		memcpy(workp, srcp, row_size);
		memcpy(workp + (height-1)*blur_pitch, srcp + (height-1)*src_pitch, row_size);
		for (y = 0; y < height; y++)
		{
			*((unsigned int *)(&workp[0])) = *((unsigned int *)(&srcp[0]));
			*((unsigned int *)(&workp[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
			srcp += src_pitch;
			workp += blur_pitch;
		}

		/* Now sharpen the edge areas and we're done! */
 		srcp = srcp_saved + src_pitch;
 		dstp = dstp_saved + blur_pitch;
		workp = workp_saved + blur_pitch;
		blurp = blurp_saved + blur_pitch;
		for (y = 1; y < height - 1; y++)
		{
			for (x = 4; x < row_size - 4; x+=4)
			{
				int xplus1 = x + 1, xplus2 = x + 2;

				if (dstp[x])
				{
					b4 = (4*(int)srcp[x] - 3*blurp[x]);
					g4 = (4*(int)srcp[x+1] - 3*blurp[x+1]);
					r4 = (4*(int)srcp[x+2] - 3*blurp[x+2]);

					if (b4 < 0) b4 = 0;
					if (g4 < 0) g4 = 0;
					if (r4 < 0) r4 = 0;
					max = b4;
					if (g4 > max) max = g4;
					if (r4 > max) max = r4;
					if (max > 255)
					{
						b4 = (b4 * 255) / max;
						g4 = (g4 * 255) / max;
						r4 = (r4 * 255) / max;
					}
					workp[x]      = (strength * b4 + invstrength * srcp[x])      >> 8;
					workp[xplus1] = (strength * g4 + invstrength * srcp[xplus1]) >> 8;
					workp[xplus2] = (strength * r4 + invstrength * srcp[xplus2]) >> 8;
				}
				else
				{
					workp[x]   = srcp[x];
					workp[xplus1] = srcp[xplus1];
					workp[xplus2] = srcp[xplus2];
				}
			}
			srcp += src_pitch;
			dstp += blur_pitch;
			workp += blur_pitch;
			blurp += blur_pitch;
		}
		return work;
	}
	else
	{
		env->ThrowError("MSharpen: Unknown color space; use YUY2, RGB32, or planar (e.g., YV12)");
		return src;		// to make compiler happy
	}
}

/**********************************
 * MMX / ISSE by Klaus Post and Milan Cutka.
 *
 * Delivers same result as YUY2, except planes
 *  are processed independently.
 *********************************/

void MSharpen::blur_plane(PVideoFrame src, PVideoFrame blur, unsigned char *blurp, int plane) {
  __int64 mask1 = 0x00001C711C711C71;
  __int64 mask2 = 0x1C711C711C710000;
  __int64 mask3 = 0x0000200000002000;
  __int64 mask4 = 0x0000000000ff0000;
  const unsigned char *srcp = src->GetReadPtr(plane);
	const unsigned char *srcp_saved = srcp;
	unsigned char *blurp_saved = blurp;
  int src_pitch = src->GetPitch(plane);
  int blur_pitch = blur->GetPitch(plane);
  int h = src->GetHeight(plane);
  int w = src->GetRowSize(plane);
  int wh = w/2;
  blurp+=blur_pitch+1;

  for (int y=1; y<h-1 ;y++) {
    __asm {
      mov esi, srcp
      mov ecx, blurp
      mov ebx, src_pitch
      mov edi, esi 
      mov edx, esi 
      add edi, ebx
      add edx, ebx
      add edx, ebx  // pitch *2
      xor eax, eax
      movq mm7,mask1
      movq mm6,mask2
     align 16
xloop:
      movd mm0,[esi+eax]
      pxor mm5,mm5
       movd mm1,[edi+eax]
      movd mm2,[edx+eax]
       punpcklbw mm0,mm5  // mm0 low
      punpcklbw mm1,mm5   // mm1 low
       movq mm3, mm0
      punpcklbw mm2,mm5   // mm2 low
       movq mm4, mm1
      movq mm5, mm2
      paddw mm0,mm1
       paddw mm3,mm4
      paddw mm0,mm2
       paddw mm3,mm5
      pmaddwd mm0,mm7
       movq mm5,[mask3]
      pmaddwd mm3,mm6
       pshufw mm1,mm0, 11101110b  // Move upper to lower
      pshufw mm4,mm3, 11101110b   
       paddd mm0,mm1
      paddd mm3,mm4
       paddd mm0, mm5
      paddd mm3, mm5
       psrld mm0,16
      pand mm3,[mask4]
      psrld mm3,8
      por mm0,mm3
      movd [eax+ecx],mm0
      add eax,2
      cmp eax,[w]
      jle xloop;
    }
    srcp += src_pitch;
    blurp += blur_pitch;
  }
	/* Fix up blur frame borders. */
	srcp = srcp_saved;
	blurp = blurp_saved;
	memcpy(blurp, srcp, w);
	memcpy(blurp + (h-1)*blur_pitch, srcp + (h-1)*src_pitch, w);
	for (y = 0; y < h; y++)
	{
		blurp[0] = srcp[0];
		blurp[w-1] = srcp[w-1];
		srcp += src_pitch;
		blurp += blur_pitch;
	}
	__asm emms;
}
void MSharpen::detect_edges(PVideoFrame src, PVideoFrame dst, unsigned char *dstp, int plane) {
  static __declspec(align(8)) const unsigned __int64 m255=0xffffffffffffffff;
  static __declspec(align(8)) unsigned  __int64 threshold64;
  threshold64=max(1,threshold);
  threshold64=threshold64+(threshold64<<8)+(threshold64<<16)+(threshold64<<24)+(threshold64<<32)+(threshold64<<40)+(threshold64<<48)+(threshold64<<56);
  const unsigned char *srcp = src->GetReadPtr(plane);
  int src_pitch = src->GetPitch(plane);
  int dst_pitch = dst->GetPitch(plane);
  const unsigned char *srcpn = srcp+src_pitch;
  int h = src->GetHeight(plane);
  int w = src->GetRowSize(plane);
	unsigned char *dstp_saved = dstp;

 for (int y=0;y<h-1;y++)
  {
   unsigned char *dstpend=dstp+w-1;
   __asm 
    {
     mov esi,[srcp]
     inc esi
     mov ecx,[srcpn]
     inc ecx
     mov edx,[srcpn]
     dec edx
     mov edi,[dstp]
     inc edi
     mov eax,[dstpend]
     movq mm6,[threshold64]
     pxor mm5,mm5
     movq mm4,[m255]
     align 16
    diag1:
     movq mm2,[ecx]
      movq mm1,[esi]
     movq mm3,[edx]

     //abs(mm1-mm2)
      movq mm7,mm1
     pminub mm1,mm2
      pmaxub mm2,mm7
     psubusb mm2,mm1
      movq mm1,[esi]
     psubusb mm2,mm6
     
     //abs(mm1-mm3)
      movq mm7,mm1
     pminub mm1,mm3
      pmaxub mm3,mm7
     psubusb mm3,mm1
     psubusb mm3,mm6

     por     mm2,mm3
     pcmpeqb mm2,mm5
     pxor    mm2,mm4
     movq    [edi],mm2
     
     add esi,8
     add ecx,8
     add edx,8
     add edi,8
     cmp edi,eax
     jl  diag1
   }
   srcp+=src_pitch;
   srcpn+=src_pitch;
   dstp+=dst_pitch;
  }
  if (mask) {
    dstp=dstp_saved;
    memset(dstp_saved+(h-1)*dst_pitch,0,w);  // Not used, if not returning mask
    for (y=0;y<h;dstp+=dst_pitch,y++) {
      dstp[0]=0;
      dstp[1]=0;
      dstp[w-1]=0;
      dstp[w-2]=0;
    }
  }
  __asm emms;
}

void MSharpen::detect_edges_HiQ(PVideoFrame src, PVideoFrame dst, unsigned char *dstp, int plane) {
  // Vertical detail detection
  const unsigned char *srcp = src->GetReadPtr(plane);
	const unsigned char *srcp_saved = srcp;
  const unsigned char *srcpn;
	unsigned char *dstp_saved = dstp;
  int src_pitch = src->GetPitch(plane);
  int dst_pitch = dst->GetPitch(plane);
  int h = src->GetHeight(plane);
  int w = src->GetRowSize(plane);

  int b1,b2;

  for (int x=0;x<w;x++)
  {
    srcp=srcp_saved;
    srcpn=srcp+src_pitch;
    dstp=dstp_saved;
    b1=srcp[x];
    for (int y=0;y<h-1;dstp+=dst_pitch,srcp+=src_pitch,srcpn+=src_pitch,y++)
    {
      b2=srcpn[x];
      if (abs(b1-b2)>=threshold)
        dstp[x]=255;
      b1=b2;
    }
  }
  
  // Horizontal detail detection
  srcp=srcp_saved;
  dstp=dstp_saved;
  for (int y=0;y<h;dstp+=dst_pitch,srcp+=src_pitch,y++)
  {
    b1=srcp[x];
    for (int x=0;x<w-1;x++)
    {
      b2=srcp[x+1];
      if (abs(b1-b2)>=threshold)
        dstp[x]=255;
      b1=b2;
    }
  }
  // Fix up detail map borders
  dstp = dstp_saved;
  memset(dstp,0,w);
  memset(dstp+dst_pitch,0,w);
  memset(dstp+(h-2)*dst_pitch,0,w);
  memset(dstp+(h-1)*dst_pitch,0,w);
  for (y=0;y<h;dstp+=dst_pitch,y++)
  {
    dstp[0]=0;
    dstp[1]=0;
    dstp[w-1]=0;
    dstp[w-2]=0;
  }
}

void MSharpen::apply_filter(PVideoFrame src, PVideoFrame blur, PVideoFrame dst, unsigned char *dstp, int plane) {
  // TODO: MMX / ISSE
  const unsigned char *srcp = src->GetReadPtr(plane);
  const unsigned char *blurp = blur->GetReadPtr(plane);
  const unsigned char *srcp_saved = srcp;
  unsigned char *dstp_saved = dstp;
  const unsigned char *blurp_saved = blurp;
  int src_pitch = src->GetPitch(plane);
  int blur_pitch = blur->GetPitch(plane);
  int dst_pitch = dst->GetPitch(plane);
  int h = src->GetHeight(plane);
  int w = src->GetRowSize(plane);
  
  memcpy(dstp,srcp,w);
  memcpy(dstp+(h-1)*dst_pitch,srcp+(h-1)*src_pitch,w);
  for (int y=0;y<h;srcp+=src_pitch,dstp+=dst_pitch,y++)
  {
    dstp[0]=srcp[0];
    dstp[w-1]=srcp[w-1];
  }
  
  // Now sharpen the edge areas and we're done
  srcp=srcp_saved+src_pitch;
  dstp=dstp_saved+dst_pitch;
  blurp=blurp+dst_pitch;
  int b4;
  for (y=1;y<h-1;srcp+=src_pitch,dstp+=dst_pitch,blurp+=blur_pitch,y++)
  {
    for (int x=1;x<w-1;)
    {
      if (*(int*)(dstp+x)==0) 
      {
        *(int*)(dstp+x)=*(int*)(srcp+x);
        x+=4;
        continue;
      }
      if (dstp[x])
      {                                     
        b4=4*int(srcp[x])-3*int(blurp[x]);
        if (b4<0) b4=0; else if (b4>255) b4=255;
        dstp[x]=(strength*b4+invstrength*srcp[x])>>8;
      }
      else
        dstp[x]=srcp[x];
      x++; 
    }  
    dstp[w-1]=srcp[w-1]; 
  }   
}


AVSValue __cdecl Create_MSharpen(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char path[1024];
	char buf[80], *p;
	int  threshold = 15;
	int	 strength = 100;
	bool highq = true;
	bool mask = false;
	bool debug = false;

	/* Load user defaults if they exist. */
	try
	{
		FILE *f;

		const char* plugin_dir = env->GetVar("$PluginDir$").AsString();
		strcpy(path, plugin_dir);
		strcat(path, "\\MSharpen.def");
		if ((f = fopen(path, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				if (strncmp(buf, "threshold=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					threshold = atoi(p);
				}
				if (strncmp(buf, "strength=", 9) == 0)
				{
					p = buf;
					while(*p++ != '=');
					strength = atoi(p);
				}
				if (strncmp(buf, "highq=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') highq = true;
					else highq = false;
				}
				if (strncmp(buf, "mask=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') mask = true;
					else mask = false;
				}
				if (strncmp(buf, "debug=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') debug = true;
					else debug = false;
				}
			}
		}
	}
	catch (...)
	{
		// plugin directory not set
		// probably using an older version avisynth
	}

    return new MSharpen(args[0].AsClip(),
						args[1].AsInt(threshold),		// edge detail detect threshold
						args[2].AsInt(strength),		// sharpening strength
						args[3].AsBool(highq),			// high quality edge detection
						args[4].AsBool(mask),			// show mask
						args[5].AsBool(debug),			// debug
						env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("MSharpen", "c[threshold]i[strength]i[highq]b[mask]b[debug]b", Create_MSharpen, 0);
    return 0;
}

